/*
 * route	This file contains an implementation of the command
 *		that manages the routing table in the kernel.
 *
 * Usage:	route [-nv]
 *		route [-v] del target
 *		route [-v] add {-net|-host} target [[gw] gateway]
 *			       [metric NN] [netmask mask]
 *
 * Version:	@(#)route.c	1.90	01/29/94
 *
 * Author:	Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *		Copyright 1993,1994 MicroWalt Corporation
 *
 *		This program is free software; you can redistribute it
 *		and/or  modify it under  the terms of  the GNU General
 *		Public  License as  published  by  the  Free  Software
 *		Foundation;  either  version 2 of the License, or  (at
 *		your option) any later version.
 */
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/socket.h>
#include <net/if.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "support.h"
#include "pathnames.h"
#include "version.h"


#define DFLT_AF	"inet"


char *Release = RELEASE,
     *Version = "@(#) route 1.90 (01/29/94)";


struct aftype *ap;
int opt_n = 0;				/* numerical output flag	*/
int opt_v = 0;				/* debugging output flag	*/
int skfd = -1;


static void
usage(void)
{
  fprintf(stderr, "Usage: route [-nv]\n");
  fprintf(stderr, "       route [-v] del target\n");
  fprintf(stderr, "       route [-v] add {-net|-host} target [[gw] gateway]\n");
  fprintf(stderr, "                  [metric NN] [netmask mask]\n");
  exit(-1);
}


static void
rt_print(void)
{
  char buff[1024], iface[16], flags[16];
  char gate_addr[128], net_addr[128];
  char mask_addr[128];
  struct sockaddr snet, sgate, smask;
  struct aftype *xap = ap;
  FILE *fp;
  int num, iflags, refcnt, use, metric;

  printf("Kernel routing table\n");
  printf(
	"Destination     Gateway         Genmask         "
	"Flags Metric Ref Use    Iface\n");
  if ((fp = fopen(_PATH_PROCNET_ROUTE, "r")) == NULL) {
	perror(_PATH_PROCNET_ROUTE);
	return;
  }

  while (!feof(fp)) {
	fgets(buff, 1023, fp);
	num = sscanf(buff, "%s %s %s %s %X %d %d %d\n",
		 iface, net_addr, gate_addr, mask_addr,
		 &iflags, &metric, &refcnt, &use);
	if (num != 8) continue;

	/* Fetch and resolve the target address. */
	(void) getsock(net_addr, &snet);
	if ((xap = get_afntype(snet.sa_family)) == NULL) {
		fprintf(stderr, "route: unsupported address family %d !\n",
							snet.sa_family);
		continue;
	}
	strcpy(net_addr, xap->sprint(&snet, (opt_n | 0x8000)));
	net_addr[15] = '\0';
    
	/* Fetch and resolve the gateway address. */
	(void) getsock(gate_addr, &sgate);
	strcpy(gate_addr, xap->sprint(&sgate, opt_n));
	gate_addr[15] = '\0';

	/* Fetch and resolve the genmask. */
	(void) getsock(mask_addr, &smask);
	strcpy(mask_addr, xap->sprint(&smask, 1));
	mask_addr[15] = '\0';

	/* Decode the flags. */
	flags[0] = '\0';
	if (iflags & RTF_UP) strcat(flags, "U");
	if (iflags & RTF_GATEWAY) strcat(flags, "G");
	if (iflags & RTF_HOST) strcat(flags, "H");
	if (iflags & RTF_REINSTATE) strcat(flags, "R");
	if (iflags & RTF_DYNAMIC) strcat(flags, "D");
	if (iflags & RTF_MODIFIED) strcat(flags, "M");

	/* Print the info. */
	printf("%-15s %-15s %-15s %-5s %-6d %-3d %6d %s\n",
		net_addr, gate_addr, mask_addr, flags,
		metric, refcnt, use, iface);
  }

  (void) fclose(fp);
}


/* Add a routing table entry. */
int
rt_add(char **args)
{
  char target[128], gateway[128];
  char netmask[128];
  struct rtentry rt;
  int xflag, isnet;

  xflag = 0;
  if (*args == NULL) usage();
  strcpy(target, *args++);
  if (!strcmp(target, "-net")) {
	xflag = 1;
	if (*args == NULL) usage();
	strcpy(target, *args++);
  } else if (!strcmp(target, "-host")) {
	xflag = 2;
	if (*args == NULL) usage();
	strcpy(target, *args++);
  }

  /* Clean out the RTREQ structure. */
  memset((char *) &rt, 0, sizeof(struct rtentry));
  if ((isnet = ap->input(target, &rt.rt_dst)) < 0) {
	ap->herror(target);
	return(-1);
  }
  switch(xflag) {
	case 1:
		isnet = 1;
		break;

	case 2:
		isnet = 0;
		break;

	default:
		break;
  }

  /* Fill in the other fields. */
  rt.rt_flags = (RTF_UP | RTF_HOST);
  if (isnet) rt.rt_flags &= ~RTF_HOST;

  /* Did we specify a GATEWAY entry? */
  if (*args != (char *)NULL) {
	if (!strcmp(*args, "gw")) args++;
  }

  if (*args != (char *)NULL) {
	strcpy(gateway, *args);
	if ((isnet = ap->input(gateway, &rt.rt_gateway)) < 0) {
		ap->herror(gateway);
		return(-1);
	}
	if (isnet) {
		fprintf(stderr,
			"%s: cannot use a NETWORK as gateway!\n",
							gateway);
		return(-1);
	}
	rt.rt_flags |= RTF_GATEWAY;
	args++;
  } else strcpy(gateway, "NONE");

  rt.rt_metric = 1;
  if ((*args != (char *)NULL) && !strcmp(*args, "metric")) {
	args++;
	rt.rt_metric = atoi(*args++);
	if (rt.rt_metric == 0) rt.rt_metric = -1;
  }

  rt.rt_genmask.sa_family = AF_UNSPEC;
  if ((*args != (char *)NULL) && !strcmp(*args, "netmask")) {
	args++;
	strcpy(netmask, *args);
	if (ap->input(netmask, &rt.rt_genmask) < 0) {
		ap->herror(netmask);
		return(-1);
	}
  }

  /* Tell the kernel to accept this route. */
  if (ioctl(skfd, SIOCADDRT, &rt) < 0) {
	fprintf(stderr, "SIOCADDRT: %s\n", strerror(errno));
	return(-1);
  }

  return(0);
}


/* Delete a routing table entry. */
int
rt_del(char **args)
{
  char target[128];
  struct sockaddr trg;
  struct rtentry rt;

  strcpy(target, *args++);
  if (ap->input(target, &trg) < 0) {
	ap->herror(target);
	return(-1);
  }

  /* Clean out the RTREQ structure. */
  memset((char *) &rt, 0, sizeof(struct rtentry));
  memcpy((char *) &rt.rt_dst, (char *) &trg, sizeof(struct sockaddr));

  /* Tell the kernel to delete this route. */
  if (ioctl(skfd, SIOCDELRT, &rt) < 0) {
	fprintf(stderr, "SIOCDELRT: %s\n", strerror(errno));
	return(-1);
  }
  return(0);
}


int
main(int argc, char **argv)
{
  int i;
  char *s;

  /* Fetch the command-line arguments. */
  argv++;
  while ((s = *argv) != NULL) {
	if (*s != '-') break;
	while (*++s != '\0') switch(*s) {
		case 'n':
			opt_n = 1;
			break;

		case 'v':
			opt_v = 1;
			break;

		default:
			usage();
	}
	argv++;
  }

  /*
   * The syntax of this commandline is getting worse
   * than ugly, a bit like that of the 'ifconfig'
   * command.  Oh well :->
   */
  ap = get_aftype(DFLT_AF);
  if (*argv != NULL) {
	if (strcmp(*argv, "add") && strcmp(*argv, "del")) {
		if ((ap = get_aftype(*argv)) == NULL) usage();
		argv++;
		if (*argv != NULL) {
			if (strcmp(*argv, "add") &&
			    strcmp(*argv, "del")) usage();
		}
	}
  }

  /* Do we have to show the contents of the routing table? */
  if (*argv == NULL) {
	rt_print();
	exit(0);
  }

  /* Create a socket to the INET kernel. */
  if ((skfd = open(_PATH_DEV_SOCKET, O_RDWR)) < 0) {
	perror("socket");
	exit(-1);
  }

  /* See what we have to do here. */
  if (!strcmp(*argv, "add")) i = rt_add(++argv);
    else i = rt_del(++argv);

  /* Close the socket. */
  (void) close(skfd);

  return(i);
}
